/**
 * Copyright (c) 2009-2012 EBM WebSourcing, 2012-2018 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.systemstate;

import java.io.File;
import java.io.InputStream;
import java.io.StringWriter;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBElement;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Marshaller;
import javax.xml.bind.Unmarshaller;
import javax.xml.transform.Source;
import javax.xml.transform.stream.StreamSource;
import javax.xml.validation.Schema;
import javax.xml.validation.SchemaFactory;

import org.ow2.petals.systemstate.generated.SystemState;
import org.xml.sax.SAXException;

/**
 * The System State builder allows to marshall/unmarshall the petalsSystemState.xml file.
 * @author Roland Naudin - EBM WebSourcing
 */
public final class SystemStateBuilder {

    /**
     * XSD schema
     */
    public static final String NAMESPACE_XSD = "http://www.w3.org/2001/XMLSchema";

    /**
     * Topology schema
     */
    public static final String SYSTEMSTATE_XSD = "petalsSystemState.xsd";

    /**
     * The unmarshaller.
     */
    private static Unmarshaller unmarshaller;

    /**
     * The exception raised during JAXB unique unmarshaller creation.
     */
    private static SystemStateException unmarshCreationEx;

    /**
     * The JAXB unique marshaller.
     */
    private static Marshaller marshaller = null;

    /**
     * The exception raised during JAXB unique marshaller creation.
     */
    private static SystemStateException marshallerCreationEx = null;

    /**
     * Static object initialization
     */
    static {

        final InputStream schemaInputStream = SystemStateBuilder.class.getResourceAsStream("/"
                + SYSTEMSTATE_XSD);

        try {
            final SchemaFactory factory = SchemaFactory.newInstance(NAMESPACE_XSD);
            final Source[] schemas = new StreamSource[] { new StreamSource(schemaInputStream) };
            final Schema schema = factory.newSchema(schemas);

            try {
                final JAXBContext jaxbContext = JAXBContext
                        .newInstance(new Class[] { SystemState.class });
                unmarshaller = jaxbContext.createUnmarshaller();
                unmarshaller.setSchema(schema);
            } catch (final JAXBException e) {
                unmarshCreationEx = new SystemStateException(
                        "Failed to create the JAXB unmarshaller", e);
            }

            // Marshaller initialization
            try {
                final JAXBContext jaxbContext = JAXBContext
                        .newInstance(new Class[] { SystemState.class });
                marshaller = jaxbContext.createMarshaller();
                marshaller.setSchema(schema);
            } catch (final JAXBException e) {
                marshallerCreationEx = new SystemStateException(
                        "Failed to create the JAXB marshaller", e);
            }

        } catch (final SAXException e) {
            final SystemStateException schemaCreationEx = new SystemStateException(
                    "Failed to parse resource '" + SystemStateBuilder.SYSTEMSTATE_XSD + "'", e);
            unmarshCreationEx = new SystemStateException("Failed to create the JAXB unmarshaller",
                    schemaCreationEx);
            marshallerCreationEx = new SystemStateException("Failed to create the JAXB marshaller",
                    schemaCreationEx);
        }
    }

    /**
     * No constructor available.
     */
    private SystemStateBuilder() {
        // NOP
    }

    /**
     * Create the system state object.
     * 
     * @param xmlFileName
     *            the xml file name
     * @return the topology
     * @throws SystemStateException
     *             impossible to create the system state
     */
    public static SystemState createSystemState(final String xmlFileName) throws SystemStateException {
        return SystemStateBuilder.createSystemState(new File(xmlFileName));
    }

    /**
     * Create the system state object.
     * 
     * @param xmlFile
     *            the xml file
     * @return the system state
     * @throws SystemStateException
     *             impossible to create the system state
     */
    public static SystemState createSystemState(final File xmlFile) throws SystemStateException {

        if (unmarshaller == null) {
            throw unmarshCreationEx;
        } else {
            try {
                final JAXBElement<SystemState> root;
                // The default Xerces unmarshaller is not thread safe
                synchronized (unmarshaller) {
                    root = unmarshaller.unmarshal(new StreamSource(xmlFile), SystemState.class);
                }

                return root.getValue();
            } catch (final JAXBException e) {
                throw new SystemStateException("Failed to build System State", e);
            }
        }
    }

    /**
     * Build the XML String from the System State in Java classes form.
     * 
     * @param systemStateClass
     *            The system state to marshall
     * @return The system state marshalled into a {@link String}.
     * @throws SystemStateException
     *             The exception raised during the marshaller creation or the
     *             exception raised during the build of the XML string.
     */
    public static String buildXmlStringSystemState(final SystemState systemStateClass)
            throws SystemStateException {

        if (marshaller == null) {
            throw marshallerCreationEx;
        } else {
            try {
                final StringWriter stringWriter = new StringWriter();
                // The default Xerces marshaller is not thread safe
                synchronized (marshaller) {
                    marshaller.marshal(systemStateClass, stringWriter);
                }

                return stringWriter.toString();
            } catch (JAXBException e) {
                throw new SystemStateException(
                        "Failed to build XML binding from System State Java classes", e);
            }
        }
    }
    
    /**
     * Build the XML File from the System State in Java classes form.
     * 
     * @param systemStateClass,
     *            The system state to marshall in the given file.
     * @param systemStateFile
     *            The file in which the system state will be marshalled.
     * @throws SystemStateException
     *             The exception raised during the marshaller creation or the
     *             exception raised during the build of the XML string.
     */
    public static void buildXmlFileSystemState(final SystemState systemStateClass, final File systemStateFile)
            throws SystemStateException {

        if (marshaller == null) {
            throw marshallerCreationEx;
        } else {
            try {
                // The default Xerces marshaller is not thread safe
                synchronized (marshaller) {
                    marshaller.marshal(systemStateClass, systemStateFile);
                }
            } catch (JAXBException e) {
                throw new SystemStateException(
                        "Failed to build XML binding from System State Java classes", e);
            }
        }
    }
}
